@@ -13,6 +13,28 @@
            tz="UTC", use.tz=TRUE)
 }
 
+.rfc2047_subject <- function(subject) {
+  subject <- enc2utf8(subject)
+  prefix <- "=?"
+  charset <- "UTF-8?"
+  encoding <- "B?"
+  suffix <- "?="
+  subject <- charToRaw(subject)
+  subject <- base64enc::base64encode(subject, linewidth = 63)
+  paste(paste0(prefix, charset, encoding, subject, suffix), collapse = " ")
+}
+
+.get_recipients <- function(headers) {
+  res <- headers$To
+  if (!is.null(headers$Cc)) {
+    res <- c(res, headers$Cc)
+  }
+  if (!is.null(headers$Bcc)) {
+    res <- c(res, headers$Bcc)
+  }
+  res
+}
+
 .write_mail <- function(headers, msg, sock) {
   if (!is.list(msg))
     msg <- list(msg)
@@ -22,8 +44,14 @@
   headers$`MIME-Version` <- "1.0"
   headers$`Content-Type` <- sprintf("multipart/mixed; boundary=\"%s\"", boundary)
 
-  writeLines(paste(names(headers),
-                   unlist(headers), sep=": "),
+  headers$To <- paste(headers$To, collapse=", ")
+  if (!is.null(headers$Cc))
+    headers$Cc <- paste(headers$Cc, collapse=", ")
+  ## Do not include BCC recipients in headers, after all, it is a
+  ## _blind_ carbon-copy.
+  headers$Bcc <- NULL
+
+  writeLines(paste(names(headers), unlist(headers), sep=": "),
              sock, sep="\r\n")
   writeLines("", sock, sep="\r\n")
 
@@ -43,12 +71,14 @@
 }
 
 .smtp_submit_mail <- function(server, port, headers, msg, verbose=FALSE) {
-  stopifnot(is.character(headers$From), is.character(headers$To))
-  
+  stopifnot(is.character(headers$From))
+
   wait_for <- function(lcode) {
     done <- FALSE
     while (!done) {
       line <- readLines(con=sock, n=1)
+      if( 0 == length(line))
+        stop("no data from server connection (this can be due to authentication issues)")
       if (verbose)
         message("<< ", line)
       code <- substring(line, 1, 3)
@@ -61,7 +91,7 @@
         else
           message("Unknown SMTP code: ", code)
       }
-      
+
     }
     return(list(code=code, msg=msg))
   }
@@ -91,7 +121,8 @@
   send_command(paste("MAIL FROM: ", headers$From), 250)
   ## >> RCPT TO: <bah@baz.org>
   ## << 250 2.1.5 Ok
-  send_command(paste("RCPT TO: ", headers$To), 250)
+  recipients <- .get_recipients(headers)
+  sapply(recipients, function(x) send_command(paste("RCPT TO: ", x), 250))
   ## >> DATA
   ## << 354 blah fu
   send_command("DATA", 354)
@@ -100,9 +131,9 @@
     message(">> <message data>")
 
   .write_mail(headers, msg, sock)
-  
+
   writeLines(".", sock, sep="\r\n")
-  
+
   wait_for(250)
   ## << 250 2.0.0 Ok: queued as XXXXXXXX
   ## >> QUIT
@@ -116,7 +147,9 @@
 ##' @title Send mail from within R
 ##'
 ##' @param from From whom the mail message is (RFC2822 style address).
-##' @param to Recipient of the message (valid RFC2822 style address).
+##' @param to Recipient of the message (vector of valid RFC2822 style addresses).
+##' @param cc Carbon-copy recipients (vector of valid RFC2822 style addresses).
+##' @param bcc Blind carbon-copy recipients (vector of valid RFC2822 style addresses).
 ##' @param subject Subject line of message.
 ##' @param msg Body text of message or a list containing
 ##'   \code{\link{mime_part}} objects.
@@ -127,7 +160,7 @@
 ##'
 ##' @seealso \code{\link{mime_part}} for a way to add attachments.
 ##' @keywords utilities
-##' 
+##'
 ##' @examples
 ##' \dontrun{
 ##' from <- sprintf("<sendmailR@@\\%s>", Sys.info()[4])
@@ -139,29 +172,38 @@
 ##' }
 ##'
 ##' @export
-sendmail <- function(from, to, subject, msg, ...,
+sendmail <- function(from, to, subject, msg, cc, bcc, ...,
                      headers=list(),
                      control=list()) {
   ## Argument checks:
   stopifnot(is.list(headers), is.list(control))
   if (length(from) != 1)
     stop("'from' must be a single address.")
-  
-  if (length(to) != 1)
-    stop("'to' must be a single address.")
-  
+
+  if (length(to) < 1)
+    stop("'to' must contain at least one address.")
+
   get_value <- function(n, default="") {
     if (n %in% names(control)) {
       return(control[[n]])
     } else if (n %in% names(.SendmailREnv$options)) {
       return(.SendmailREnv$options[[n]])
     } else {
-      return(default)      
+      return(default)
     }
   }
-  
+
   headers$From <- from
   headers$To <- to
+  if (!missing(cc))
+    headers$Cc <- cc
+  if (!missing(bcc))
+    headers$Bcc <- bcc
+
+  ## Encode subject if it contains none ASCII characters
+  if (grepl("[^ -~]", subject)) {
+    subject <- .rfc2047_subject(subject)
+  }
   headers$Subject <- subject
 
   ## Add Date header if not explicitly set. This fixes the annoyance,
@@ -169,15 +211,16 @@ sendmail <- function(from, to, subject, msg, ...,
   ## do not have a Date header.
   if (is.null(headers$Date))
     headers$Date <- .rfc2822_date()
-  
+
   transport <- get_value("transport", "smtp")
   verbose <- get_value("verbose", FALSE)
   if (transport == "smtp") {
     server <- get_value("smtpServer", "localhost")
     port <- get_value("smtpPort", 25)
-    
+
     .smtp_submit_mail(server, port, headers, msg, verbose)
   } else if (transport == "debug") {
-    .write_mail(headers, msg, stdout())
+    message("Recipients: ", paste(.get_recipients(headers), collapse=", "))
+    .write_mail(headers, msg, stderr())
   }
 }
